/* Boost interval/detail/x86gcc_rounding_control.hpp file
 *
 * Copyright 2000 Jens Maurer
 * Copyright 2002 Herv Brnnimann, Guillaume Melquiond, Sylvain Pion
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or
 * copy at http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef BOOST_NUMERIC_INTERVAL_DETAIL_X86GCC_ROUNDING_CONTROL_HPP
#define BOOST_NUMERIC_INTERVAL_DETAIL_X86GCC_ROUNDING_CONTROL_HPP

#ifndef __GNUC__
#  error This header only works with GNU CC.
#endif

// #ifndef __i386__
// #  error This header only works on x86 CPUs.
// #endif

// #if defined(__x86_64__) || defined(__SSE2_MATH__)
// #include <xmmintrin.h>
// #endif

namespace boost {
namespace numeric {
namespace interval_lib {
namespace detail {

struct x86_rounding
{
  typedef unsigned short rounding_mode;

  static void set_rounding_mode(const rounding_mode& mode)
  {
#ifdef __i386__
    __asm__ __volatile__ ("fldcw %0" : : "m"(mode));
#endif
#if defined(__x86_64__) || defined(__SSE2_MATH__)
    // _mm_setcsr( ((mode & 0xc00) << 3) | ( (mode & 0x3f) << 7) );
    unsigned int tmp;
    tmp = ((mode & 0xc00) << 3) | ((mode & 0x3f) << 7);
    __asm__ __volatile__ ("ldmxcsr %0" : : "m"(tmp));
#endif
  }

  static void get_rounding_mode(rounding_mode& mode)
  {
#ifdef __i386
    __asm__ __volatile__ ("fnstcw %0" : "=m"(mode));
#else
    unsigned int tmp;
    // tmp = _mm_getcsr();
    __asm__ __volatile__ ("stmxcsr %0" : "=m"(tmp));
    mode = ((tmp & 0x6000) >> 3) | ((tmp & 0x1f80) >> 7);
#endif
  }

  template<class T>
  static T to_int(T r)
  {
    return rint(r);
    #if 0
    T r_;
    __asm__ ("frndint" : "=&t"(r_) : "0"(r));
    return r_;
    #endif
  }
};

} // namespace detail
} // namespace interval_lib
} // namespace numeric
} // namespace boost

#endif /* BOOST_NUMERIC_INTERVAL_DETAIL_X86GCC_ROUNDING_CONTROL_HPP */
